@@ -56,11 +56,7 @@ int tty_reopen()
 {
   logw(3, "tty re-opening...");
   tty_close(&tty);
-#ifdef  TRXCTL
-  tty_init(&tty, cfg.ttyport, cfg.ttyspeed, cfg.trxcntl);
-#else
-  tty_init(&tty, cfg.ttyport, cfg.ttyspeed);
-#endif
+  tty_init(&tty);
   if (tty_open(&tty) != RC_OK)
   {
 #ifdef LOG
@@ -96,11 +92,7 @@ int
 conn_init(void)
 {
   /* tty device initialization */
-#ifdef  TRXCTL
-  tty_init(&tty, cfg.ttyport, cfg.ttyspeed, cfg.trxcntl);
-#else
-  tty_init(&tty, cfg.ttyport, cfg.ttyspeed);
-#endif
+  tty_init(&tty);
   if (tty_open(&tty) != RC_OK)
   {
 #ifdef LOG
@@ -113,13 +105,12 @@ conn_init(void)
   state_tty_set(&tty, TTY_PAUSE);
 
   /* create server socket */
-  if ((server_sd =
-         sock_create_server("", cfg.serverport, TRUE)) < 0)
+  if ((server_sd = sock_create_server(cfg.serveraddr, cfg.serverport, TRUE)) < 0)
   {
 #ifdef LOG
     logw(0, "conn_init():"
-           " can't create listen() socket (%s)",
-           strerror(errno));
+           " can't create listen socket (%s)",
+           (errno != 0) ? strerror(errno) : "failed");
 #endif
     return RC_ERR;
   }
@@ -140,26 +131,28 @@ conn_open(void)
 {
   int sd;
   conn_t *newconn;
-  struct sockaddr_in rmt_addr;
+  struct sockaddr_storage rmt_addr;
+  char ipstr[INET6_ADDRSTRLEN];
 
-  if ((sd = sock_accept(server_sd, &rmt_addr, TRUE)) == RC_ERR)
+  if ((sd = sock_accept(server_sd, (struct sockaddr *)&rmt_addr,
+                        sizeof(rmt_addr), TRUE)) == RC_ERR)
   { /* error in conn_accept() */
 #ifdef LOG
     logw(0, "conn_open(): error in accept() (%s)", strerror(errno));
 #endif
     return;
   }
+ inet_ntop(rmt_addr.ss_family, sock_addr((struct sockaddr *)&rmt_addr),
+           ipstr, sizeof(ipstr));
 #ifdef LOG
-  logw(2, "conn_open(): accepting connection from %s",
-         inet_ntoa(rmt_addr.sin_addr));
+  logw(2, "conn_open(): accepting connection from %s", ipstr);
 #endif
   /* compare descriptor of connection with FD_SETSIZE */
   if (sd >= FD_SETSIZE)
   {
 #ifdef LOG
     logw(1, "conn_open(): FD_SETSIZE limit reached,"
-           " connection from %s will be dropped",
-           inet_ntoa(rmt_addr.sin_addr));
+           " connection from %s will be dropped", ipstr);
 #endif
     close(sd);
     return;
@@ -169,8 +162,7 @@ conn_open(void)
   {
 #ifdef LOG
     logw(1, "conn_open(): number of connections limit reached,"
-           " connection from %s will be dropped",
-           inet_ntoa(rmt_addr.sin_addr));
+           " connection from %s will be dropped", ipstr);
 #endif
     close(sd);
     return;
@@ -178,8 +170,7 @@ conn_open(void)
   /* enqueue connection */
   newconn = queue_new_elem(&queue);
   newconn->sd = sd;
-  memcpy((void *) &newconn->sockaddr,
-         &rmt_addr, sizeof(struct sockaddr_in));
+  memcpy((void *) &newconn->remote_addr, &ipstr, sizeof(ipstr));
   state_conn_set(newconn, CONN_HEADER);
 }
 
@@ -193,8 +184,7 @@ conn_close(conn_t *conn)
 {
   conn_t *nextconn;
 #ifdef LOG
-  logw(2, "conn_close(): closing connection from %s",
-         inet_ntoa(conn->sockaddr.sin_addr));
+  logw(2, "conn_close(): closing connection from %s", conn->remote_addr);
 #endif
   /* close socket */
   close(conn->sd);
@@ -256,43 +246,30 @@ conn_write(int d, void *buf, size_t nbytes, int istty)
   fd_set fs;
   struct timeval ts, tts;
   long delay;
+
 #ifdef TRXCTL
-  if (istty) {
-    if (cfg.trxcntl != TRX_ADDC )
-      tty_set_rts(d);
-    usleep(35000000l/cfg.ttyspeed);
+  if (istty && cfg.trxcntl != TRX_ADDC)
+  {
+    tty_set_rts(d);
+    tty_delay(35000000l/cfg.ttyspeed);
   }
 #endif
+
   FD_ZERO(&fs);
   FD_SET(d, &fs);
   do
-  { /* trying write to descriptor while breaked by signals */
-    gettimeofday(&ts, NULL);
+  {
     rc = write(d, buf, nbytes);
   } while (rc == -1 && errno == EINTR);
-  gettimeofday(&ts, NULL);
-
 
 #ifdef TRXCTL
-  if (istty) {
-#if 1
-    do {
-      gettimeofday(&tts, NULL);
-      delay = DV(nbytes, cfg.ttyspeed) -
-        ((tts.tv_sec * 1000000l + tts.tv_usec) - (ts.tv_sec * 1000000l + ts.tv_usec));
-    } while (delay > 0);
-#else
-    gettimeofday(&tts, NULL);
-    delay = DV(nbytes, cfg.ttyspeed) -
-      ((tts.tv_sec * 1000000l + tts.tv_usec) - (ts.tv_sec * 1000000l + ts.tv_usec));
-    usleep(delay);
-#endif
-/*    tcdrain(d); - hangs sometimes, so make calculated delay */
-    if (cfg.trxcntl != TRX_ADDC ) {
-      tty_clr_rts(d);
-    }
+  if (istty && cfg.trxcntl != TRX_ADDC )
+  {
+    tty_delay(DV(nbytes, tty.bpc, cfg.ttyspeed));
+    tty_clr_rts(d);
   }
 #endif
+
   return (rc < 0) ? RC_ERR : rc;
 }
 
@@ -491,8 +468,10 @@ conn_loop(void)
 #endif
                 if (!tty.trynum) {
                   modbus_ex_write(actconn->buf, MB_EX_CRC);
+#ifdef DEBUG
                   logw(3, "tty: response is incorrect (%d of %d bytes, offset %d), return error", tty.ptrbuf,
                     tty.rxoffset + tty.rxlen, tty.rxoffset);
+#endif
                 } else
                 { /* retry request */
 #ifdef DEBUG
@@ -536,15 +515,13 @@ conn_loop(void)
             if (curconn->state == CONN_TTY)
             { /* deadlock in CONN_TTY state, make attempt to reinitialize serial port */
 #ifdef LOG
-              logw(0, "conn[%s]: state CONN_TTY deadlock.",
-                     inet_ntoa(curconn->sockaddr.sin_addr));
+              logw(0, "conn[%s]: state CONN_TTY deadlock.", curconn->remote_addr);
 #endif
               tty_reinit();
             }
             /* purge connection */
 #ifdef LOG
-            logw(2, "conn[%s]: timeout, closing connection",
-                   inet_ntoa(curconn->sockaddr.sin_addr));
+            logw(2, "conn[%s]: timeout, closing connection", curconn->remote_addr);
 #endif
             curconn = conn_close(curconn);
             continue;
@@ -581,34 +558,46 @@ conn_loop(void)
 #ifdef DEBUG
           logw(7, "tty: request written (total %d bytes)", tty.txlen);
 #endif
-          state_tty_set(&tty, TTY_RESP);
-          switch (tty.txbuf[1]) {
-            case 1:
-            case 2:
-              tty.rxlen = 5 + (tty.txbuf[4] * 256 + tty.txbuf[5] + 7)/8;
-              break;
-            case 3:
-            case 4:
-              tty.rxlen = 5 + tty.txbuf[5] * 2;
-              break;
-            case 7:
-              tty.rxlen = 5;
-              break;
-            case 11:
-            case 15:
-            case 16:
-              tty.rxlen = 8;
-              break;
-            default:
-              tty.rxlen = tty.txlen;
-              break;
-          }
-          if (tty.rxlen > TTY_BUFSIZE)
-            tty.rxlen = TTY_BUFSIZE;
-          tty.timer += DV(tty.rxlen, tty.speed);
+          if (!tty.txbuf[0])
+          {
+            /* broadcast request sent, no reply expected */
+            state_conn_set(actconn, CONN_HEADER);
+            state_tty_set(&tty, TTY_PAUSE);
 #ifdef DEBUG
-          logw(5, "tty: estimated %d bytes, waiting %lu usec", tty.rxlen, tty.timer);
+            logw(5, "conn[%s]: broadcast request sent", curconn->remote_addr);
 #endif
+          }
+          else
+          {
+            state_tty_set(&tty, TTY_RESP);
+            switch (tty.txbuf[1]) {
+              case 1:
+              case 2:
+                tty.rxlen = 5 + (tty.txbuf[4] * 256 + tty.txbuf[5] + 7)/8;
+                break;
+              case 3:
+              case 4:
+                tty.rxlen = 5 + tty.txbuf[5] * 2;
+                break;
+              case 7:
+                tty.rxlen = 5;
+                break;
+              case 11:
+              case 15:
+              case 16:
+                tty.rxlen = 8;
+                break;
+              default:
+                tty.rxlen = tty.txlen;
+                break;
+            }
+            if (tty.rxlen > TTY_BUFSIZE)
+              tty.rxlen = TTY_BUFSIZE;
+            tty.timer += DV(tty.rxlen, tty.bpc, tty.speed);
+#ifdef DEBUG
+            logw(5, "tty: estimated %d bytes, waiting %lu usec", tty.rxlen, tty.timer);
+#endif
+          }
         }
       }
 
@@ -790,8 +779,7 @@ conn_loop(void)
                 /* check request function code */
                 unsigned char fc = MB_FRAME(curconn->buf, MB_FCODE);
 #ifdef DEBUG
-                logw(7, "conn[%s]: read request fc %d",
-                     inet_ntoa(curconn->sockaddr.sin_addr), fc);
+                logw(7, "conn[%s]: read request fc %d", curconn->remote_addr, fc);
 #endif
                 switch (fc)
                 {
@@ -846,11 +834,12 @@ conn_loop(void)
               { /* ### frame received completely ### */
 #ifdef DEBUG
                 t[0] = '\0';
+                int i;
                 for (i = MB_UNIT_ID; i < curconn->ctr; i++) {
                   sprintf(v, "[%2.2x]", curconn->buf[i]);
                   strncat(t, v, 1024-strlen(t));
                 }
-                logw(5, "conn[%s]: request: %s", inet_ntoa(curconn->sockaddr.sin_addr), t);
+                logw(5, "conn[%s]: request: %s", curconn->remote_addr, t);
 #endif
                 state_conn_set(curconn, CONN_TTY);
                 if (tty.state == TTY_READY)
